import logging

from .BitBucketRepoSession import BitBucketRepoSession
from .FeedRepoSession import FeedRepoSession
from .GitHubRepoSession import GitHubRepoSession
from .HelmChartRepoSession import HelmChartRepoSession
from .GitLabRepoSession import GitLabRepoSession
from .LocalVersionSession import LocalVersionSession
from .MercurialRepoSession import MercurialRepoSession
from .PypiRepoSession import PypiRepoSession
from .SourceForgeRepoSession import SourceForgeRepoSession
from .SystemRepoSession import SystemRepoSession
from .WikipediaRepoSession import WikipediaRepoSession
from .WordPressPluginRepoSession import WordPressPluginRepoSession
from .utils import BadProjectError

log = logging.getLogger(__name__)


class HolderFactory:
    HOLDERS = {
        'github': GitHubRepoSession,
        'gitlab': GitLabRepoSession,
        'bitbucket': BitBucketRepoSession,
        'pip': PypiRepoSession,
        'hg': MercurialRepoSession,
        'sf': SourceForgeRepoSession,
        'website-feed': FeedRepoSession,
        'local': LocalVersionSession,
        'helm_chart': HelmChartRepoSession,
        'wiki': WikipediaRepoSession,
        'system': SystemRepoSession,
        'wp': WordPressPluginRepoSession,
    }

    @staticmethod
    def guess_from_homepage(repo, hostname):
        # repo auto-discovery failed for detected/default provider
        # now we simply try website provider based on the hostname/RSS feeds in HTML or GitHub links
        holder = FeedRepoSession(repo, hostname)
        if not holder.is_valid():
            # re-use soup from feed holder object
            log.info('Have not found any RSS feed for the website {}'.format(hostname))
            github_link = holder.home_soup.select_one("a[href*='github.com']")
            if github_link:
                hostname, repo = GitHubRepoSession.get_host_repo_for_link(github_link['href'])
                holder = GitHubRepoSession(repo, hostname)
        return holder

    @staticmethod
    # go through subclasses in order to find the one that is holding a given project
    # repo is either complete URL or a name allowing to identify a single project
    def get_instance_for_repo(repo, at=None):
        """Find the right hosting for this repo."""
        if at == 'helm_chart' or (at and '/' not in repo):
            return HolderFactory.HOLDERS[at](repo, hostname=None)
        holder_class = HolderFactory.HOLDERS['github']
        hostname = None
        known_repo = None

        for k, sc in HolderFactory.HOLDERS.items():
            known_repo = sc.is_official_for_repo(repo)
            if known_repo:
                holder_class = sc
                log.info('Trying {} adapter'.format(k))
                break
            # TODO now easy multiple default hostnames per holder
            hostname = sc.get_matching_hostname(repo)
            if hostname:
                holder_class = sc
                break
        if known_repo:
            repo = known_repo['repo']
            # known repo tells us hosted domain of e.g. mercurial web
            if 'hostname' in known_repo:
                hostname = known_repo['hostname']
        else:
            hostname, repo = holder_class.get_host_repo_for_link(repo)

        holder = holder_class(repo, hostname)
        if not holder.is_valid() and hostname:
            holder = HolderFactory.guess_from_homepage(repo, hostname)
            if not holder.is_valid():
                raise BadProjectError(
                    'No project found. Could not guess a repo from homepage'
                )
            
        if known_repo and 'branches' in known_repo:
            holder.set_branches(known_repo['branches'])
            
        if known_repo and 'only' in known_repo:
            holder.set_only(known_repo['only'])

        if known_repo and 'release_url_format' in known_repo:
            holder.RELEASE_URL_FORMAT = known_repo['release_url_format']

        return holder



"""
lastversion
==========
License: BSD, see LICENSE for more details.
"""

__author__ = "Danila Vershinin"

import logging

from .__about__ import (
    __version__,
)
# We intentially import for export here, so it is ok to silence DeepSource test
# skipcq: PY-W2000
from .lastversion import __self__
# skipcq: PY-W2000
from .lastversion import check_version
# skipcq: PY-W2000
from .lastversion import has_update
# skipcq: PY-W2000
from .lastversion import latest
# skipcq: PY-W2000
from .lastversion import main

# https://realpython.com/python-logging-source-code/#library-vs-application-logging-what-is-nullhandler
# when used as library, we default to opt-in approach, whereas library user have to enable logging
# from lastversion
logging.getLogger(__name__).addHandler(logging.NullHandler())
# patch up https://github.com/ionrock/cachecontrol/issues/230
logging.getLogger('cachecontrol.controller').addHandler(logging.NullHandler())
logging.getLogger('pip._vendor.cachecontrol.controller').addHandler(logging.NullHandler())



import sys as _sys
from argparse import SUPPRESS, Action

import lastversion
from .__about__ import __version__


class VersionAction(Action):

    def __init__(self,
                 option_strings,
                 version=None,
                 dest=SUPPRESS,
                 default=SUPPRESS,
                 help="show program's version number and exit"):
        super(VersionAction, self).__init__(
            option_strings=option_strings,
            dest=dest,
            default=default,
            nargs=0,
            help=help)
        self.version = version

    def __call__(self, parser, namespace, values, option_string=None):
        version = '%(prog)s {version}'.format(version=__version__)
        last_version = lastversion.latest(lastversion.__self__)
        if __version__ == str(last_version):
            version += ', up to date'
        else:
            version += ', newer version {} available'.format(last_version)
        formatter = parser._get_formatter()
        formatter.add_text(version)
        parser._print_message(formatter.format_help(), _sys.stdout)
        parser.exit()



import json
import logging
import math
import os
import re
import time
from datetime import datetime
from datetime import timedelta
import feedparser
from dateutil import parser

from .ProjectHolder import ProjectHolder
from .utils import ApiCredentialsError, BadProjectError

log = logging.getLogger(__name__)

TOKEN_PRO_TIP = 'ProTip: set GITHUB_API_TOKEN env var as per ' \
                'https://github.com/dvershinin/lastversion#tips'


def set_matching_from_tag(ret, tag):
    ret = tag


def asset_matches(asset, search, regex_matching):
    """Check if the asset equals to string or satisfies a regular expression
    Args:
        asset (dict): asset dict as returned by the API
        search (str): string or regexp to match asset's name or label with
        regex_matching (bool): whether search argument is a regexp
    Returns:
        bool: Whether match is satisfied
    """
    if regex_matching:
        if asset['label'] and re.search(search, asset['label']):
            return True
        if asset['name'] and re.search(search, asset['name']):
            return True
    elif search in (asset['label'], asset['name']):
        return True
    return False


class GitHubRepoSession(ProjectHolder):
    """A class to represent a GitHub project holder."""

    DEFAULT_HOSTNAME = 'github.com'
    DEFAULT_HOLDER = True

    # one-word aliases or simply known popular repos to skip using search API
    KNOWN_REPOS_BY_NAME = {
        'php': {
            'repo': 'php/php-src',
            # get URL from official website because it is a "prepared" source
            'release_url_format': "https://www.php.net/distributions/php-{version}.tar.gz"
        },
        'linux': {'repo': 'torvalds/linux'},
        'kernel': {'repo': 'torvalds/linux'},
        'openssl': {'repo': 'openssl/openssl'},
        'python': {'repo': 'python/cpython'},
        'cmake': {'repo': 'kitware/cmake'},
        'kodi': {'repo': 'xbmc/xbmc'}
    }

    """
    The last alphanumeric after digits is part of version scheme, not beta level.
    E.g. 1.1.1b is not beta. Hard-coding such odds repos is required.
    """
    LAST_CHAR_FIX_REQUIRED_ON = [
        'openssl/openssl'
    ]

    """ The following format will benefit from:
    1) not using API, so is not subject to its rate limits
    2) likely has been accessed by someone in CDN and thus faster
    3) provides more or less unique filenames once the stuff is downloaded
    See https://fedoraproject.org/wiki/Packaging:SourceURL#Git_Tags
    We use variation of this: it does not need a parsed version (thus works for --pre better)
    and it is not broken on fancy release tags like v1.2.3-stable
    https://github.com/OWNER/PROJECT/archive/%{gittag}/%{gittag}-%{version}.tar.gz
    """
    RELEASE_URL_FORMAT = "https://{hostname}/{repo}/archive/{tag}/{name}-{tag}.{ext}"
    SHORT_RELEASE_URL_FORMAT = "https://{hostname}/{repo}/archive/{tag}.{ext}"

    def find_repo_by_name_only(self, repo):
        if repo.startswith(('https://', 'http://')):
            return None
        cache_repo_names_file = "{}/repos.json".format(self.cache_dir)
        try:
            with open(cache_repo_names_file, 'r') as reader:
                cache = json.load(reader)
        except (IOError, ValueError):
            cache = {}
        try:
            if repo in cache and time.time() - cache[repo]['updated_at'] < 3600 * 24 * 30:
                log.info("Found {} in repo short name cache".format(repo))
                if not cache[repo]['repo']:
                    raise BadProjectError(
                        'No project found on GitHub for search query: {}'.format(repo)
                    )
                return cache[repo]['repo']
        except TypeError:
            pass
        log.info("Making query against GitHub API to search repo {}".format(repo))
        r = self.get(
            '{}/search/repositories?q={}+in:name'.format(self.api_base, repo))
        if r.status_code == 404:
            # when not found, skip using this holder in the factory by not setting self.repo
            return None
        if r.status_code != 200:
            raise BadProjectError(
                'Error while identifying full repository on GitHub for search query: {}'.format(
                    repo
                )
            )
        data = r.json()
        full_name = ''
        if data['items']:
            full_name = data['items'][0]['full_name']
        cache[repo] = {
            'repo': full_name,
            'updated_at': int(time.time())
        }
        try:
            with open(cache_repo_names_file, 'w') as writer:
                writer.write(
                    json.dumps(cache)
                )
        except (IOError, ValueError):
            pass
        if not full_name:
            raise BadProjectError(
                'No project found on GitHub for search query: {}'.format(repo)
            )
        return full_name

    def __init__(self, repo, hostname):
        super(GitHubRepoSession, self).__init__()
        # dict holding repo/owner to feed contents of releases atom
        self.feed_contents = {}
        self.rate_limited_count = 0
        self.api_token = os.getenv("GITHUB_API_TOKEN")
        if not self.api_token:
            self.api_token = os.getenv("GITHUB_TOKEN")
        self.hostname = hostname
        if not self.hostname:
            self.hostname = self.DEFAULT_HOSTNAME
        # Explicitly specify the API version that we want:
        self.headers.update({
            'Accept': 'application/vnd.github.v3+json'
        })
        if self.api_token:
            log.info('Using API token.')
            self.headers.update({'Authorization': "token {}".format(self.api_token)})
        if self.hostname != self.DEFAULT_HOSTNAME:
            self.api_base = 'https://{}/api/v3'.format(self.hostname)
        else:
            self.api_base = 'https://api.{}'.format(self.DEFAULT_HOSTNAME)
        if '/' not in repo:
            official_repo = self.try_get_official(repo)
            if official_repo:
                repo = official_repo
                log.info('Using official repo {}'.format(repo))
            else:
                repo = self.find_repo_by_name_only(repo)
                if repo:
                    log.info('Using repo {} obtained from search API'.format(self.repo))
                else:
                    return
        self.set_repo(repo)

    def get_rate_limit_url(self):
        return '{}/rate_limit'.format(self.api_base)

    def get(self, url, **kwargs):
        """Send GET reqiest and account for GitHub rate limits and such."""
        r = super(GitHubRepoSession, self).get(url, **kwargs)
        log.info('Got HTTP status code {} from {}'.format(r.status_code, url))
        if r.status_code == 401:
            if self.api_token:
                raise ApiCredentialsError('API request was denied despite using an API token. '
                                          'Missing scopes?')
            raise ApiCredentialsError('Denied API access. Please set GITHUB_API_TOKEN env var '
                                      'as per https://github.com/dvershinin/lastversion#tips')
        if r.status_code == 403:
            if 'X-RateLimit-Reset' in r.headers and 'X-RateLimit-Remaining' in r.headers:
                if self.rate_limited_count > 2:
                    raise ApiCredentialsError(
                        'API requests were denied after retrying {} times'.format(
                            self.rate_limited_count)
                    )
                remaining = int(r.headers['X-RateLimit-Remaining'])
                # 1 sec to account for skewed clock between GitHub and client
                wait_for = float(r.headers['X-RateLimit-Reset']) - time.time() + 1.0
                wait_for = math.ceil(wait_for)
                if not remaining:
                    # got 403, likely due to used quota
                    if wait_for < 300:
                        if wait_for < 0:
                            log.warning(
                                'Exceeded API quota. Repeating request because quota is about to '
                                'be reinstated'
                            )
                        else:
                            w = 'Waiting {} seconds for API quota reinstatement.'.format(wait_for)
                            if "GITHUB_API_TOKEN" not in os.environ \
                                    and 'GITHUB_TOKEN' not in os.environ:
                                w = "{} {}".format(w, TOKEN_PRO_TIP)
                            log.warning(w)
                            time.sleep(wait_for)
                        self.rate_limited_count = self.rate_limited_count + 1
                        return self.get(url)
                    raise ApiCredentialsError(
                        'Exceeded API rate limit after waiting: {}'.format(
                            r.json()['message'])
                    )
                return self.get(url)

        if r.status_code == 403 and url != self.get_rate_limit_url():
            self.rate_limited_count = 0
        return r

    def rate_limit(self):
        url = '{}/rate_limit'.format(self.api_base)
        return self.get(url)

    def repo_query(self, uri):
        url = '{}/repos/{}{}'.format(self.api_base, self.repo, uri)
        return self.get(url)

    def repo_license(self, tag):
        r = self.repo_query('/license?ref={}'.format(tag))
        if r.status_code == 200:
            # unfortunately, unlike /readme, API always returns *latest* license, ignoring tag
            # we have to double check whether the license file exists "at release tag"
            license_data = r.json()
            license_path = license_data['path']
            license_r = self.repo_query('/contents/{}?ref={}'.format(license_path, tag))
            if license_r.status_code == 200:
                return license_data
        return None

    def repo_readme(self, tag):
        r = self.repo_query('/readme?ref={}'.format(tag))
        if r.status_code == 200:
            return r.json()
        return None

    def find_in_tags_via_graphql(self, ret, pre_ok, major):
        """GraphQL allows for faster search across many tags.
        We aggregate the highest semantic version among batches of 100 records.
        In this way --major filtering results in much fewer requests compared to traditional API
        use.

        Args:
            ret (dict): currently selected release object
            pre_ok (bool): whether betas are acceptable
            major (str): the major filter

        Returns: currently selected release object

        """
        query_fmt = """
        {
          rateLimit {
            cost
            remaining
          }
          repository(owner: "%s", name: "%s") {
            tags: refs(refPrefix: "refs/tags/", first: 100, after: "%s",
              orderBy: {field: TAG_COMMIT_DATE, direction: DESC}) {
              edges {
                cursor,
                node {
                  ...refInfo
                }
              }
            }
          }
        }

        fragment refInfo on Ref {
          name
          target {
            sha: oid
            commitResourcePath
            __typename
            ... on Tag {
              target {
                ... on Commit {
                  ...commitInfo
                }
              }
              tagger {
                name
                email
                date
              }
            }
            ... on Commit {
              ...commitInfo
            }
          }
        }

        fragment commitInfo on Commit {
          zipballUrl
          tarballUrl
          author {
            name
            email
            date
          }
        }

        """
        cursor = ''
        log.info('Using graphql queries...')
        while True:
            # testing on php/php-src
            owner, name = self.repo.split('/')
            query = query_fmt % (owner, name, cursor)
            log.info('Running query {}'.format(query))
            r = self.post('{}/graphql'.format(self.api_base), json={'query': query})
            log.info('Requested graphql with cursor "{}"'.format(cursor))
            if r.status_code != 200:
                log.info("query returned non 200 response code {}".format(r.status_code))
                return ret
            j = r.json()
            if 'errors' in j and j['errors'][0]['type'] == 'NOT_FOUND':
                raise BadProjectError(
                    'No such project found on GitHub: {}'.format(self.repo)
                )
            if not j['data']['repository']['tags']['edges']:
                log.info('No tags in GraphQL response: {}'.format(r.text))
                break
            for edge in j['data']['repository']['tags']['edges']:
                node = edge['node']
                cursor = edge['cursor']
                tag_name = node['name']
                version = self.sanitize_version(tag_name, pre_ok, major)
                if not version:
                    continue
                if 'tagger' in node['target']:
                    # use date of annotated tag as it better corresponds to "release date"
                    log.info('Using annotated tag date')
                    d = node['target']['tagger']['date']
                else:
                    # using commit date because the tag is not annotated
                    log.info('Using commit date')
                    d = node['target']['author']['date']
                tag_date = parser.parse(d)
                if ret and tag_date + timedelta(days=365) < ret['tag_date']:
                    log.info('The version {} is newer, but is too old!'.format(version))
                    break
                if not ret or version > ret['version'] or tag_date > ret['tag_date']:
                    # we always want to return formal release if it exists, cause it has useful
                    # data grab formal release via APi to check for pre-release mark
                    formal_release = self.get_formal_release_for_tag(tag_name)
                    if formal_release:
                        ret = self.set_matching_formal_release(ret, formal_release, version, pre_ok)
                    else:
                        if not self.having_asset:
                            ret = {
                                'tag_name': tag_name,
                                'tag_date': tag_date,
                                'version': version,
                                'type': 'graphql'
                            }
            if ret:
                break
        return ret

    def get_formal_release_for_tag(self, tag):
        r = self.repo_query('/releases/tags/{}'.format(tag))
        if r.status_code == 200:
            # noinspection SpellCheckingInspection
            return r.json()
        return None

    # finding in tags requires paging through ALL of them, because the API does not list them
    # in order of recency, thus this is very slow
    # in: current release to be returned, output: newer release to be returned
    def find_in_tags(self, ret, pre_ok, major):
        r = self.repo_query('/tags?per_page=100')
        if r.status_code != 200:
            return None
        tags = r.json()
        while 'next' in r.links.keys():
            r = self.get(r.links['next']['url'])
            tags.extend(r.json())

        for t in tags:
            tag_name = t['name']
            version = self.sanitize_version(tag_name, pre_ok, major)
            if not version:
                continue
            c = self.repo_query('/git/commits/{}'.format(t['commit']['sha']))
            d = c.json()['committer']['date']
            d = parser.parse(d)

            if not ret or version > ret['version'] or d > ret['tag_date']:
                # rare case: if upstream filed formal pre-release that passes as stable
                # version (tag is 1.2.3 instead of 1.2.3b) double check if pre-release
                # TODO handle API failure here as it may result in "false positive"?
                release_for_tag = self.get_formal_release_for_tag(tag_name)
                if release_for_tag:
                    ret = self.set_matching_formal_release(ret, release_for_tag, version, pre_ok)
                else:
                    ret = t
                    ret['tag_name'] = tag_name
                    ret['tag_date'] = d
                    ret['version'] = version
                    ret['type'] = 'tag'
        return ret

    def get_releases_feed_contents(self, rename_checked=False):
        """
        Fetch contents of repository's releases.atom feed.
        # The releases.atom and tags.atom don't differ much except releases having more data.
        # The releases.atom feed includes non-formal releases which are just tags, so we are good.
        # Based on testing, edited old releases don't jump forward in the list and stay behind (good).
        # The only downside is they don't bear pre-release mark (unlike API), and have limited data.
        # We work around these by checking pre-release flag and get full release data via API.
        """
        if self.repo in self.feed_contents:
            return self.feed_contents[self.repo]
        headers = {
            'Accept': '*/*',
            # private repos do not have releases.atom to begin with,
            # authorization header may cause a false positive 200 response with an empty feed!
            'Authorization': ''
        }
        r = self.get('https://{}/{}/releases.atom'.format(self.hostname, self.repo), headers=headers)
        # API requests are varied by cookie, we don't want serializer for cache fail because of that
        self.cookies.clear()
        if r.status_code == 404 and not rename_checked:
            # #44: in some network locations, GitHub returns 404 (as opposed to 301 redirect) for the renamed
            # repositories /releases.atom. When we get a 404, we lazily load repo info via API, and hopefully
            # get redirect there as well as the new repo full name
            r = self.repo_query('')
            if r.status_code == 200:
                repo_data = r.json()
                if self.repo != repo_data['full_name']:
                    log.info('Detected name change from {} to {}'.format(self.repo,
                                                                         repo_data['full_name']))
                    self.set_repo(repo_data['full_name'])
                    # request the feed from the new location
                    return self.get_releases_feed_contents(rename_checked=False)
        if r.status_code == 200:
            self.feed_contents[self.repo] = r.text
            return r.text
        return None

    def get_releases_feed_entries(self):
        """Get an array of releases.atom feed entries."""
        feed_contents = self.get_releases_feed_contents()
        if not feed_contents:
            log.info('The releases.atom feed failed to be fetched!')
            return None
        feed = feedparser.parse(feed_contents)
        if 'bozo' in feed and feed['bozo'] == 1 and 'bozo_exception' in feed:
            exc = feed.bozo_exception
            log.info("Failed to parse feed: {}".format(exc.getMessage()))
            return None
        if not feed.entries:
            log.info('Feed has no elements. Means no tags and no releases')
            return []
        return feed.entries

    def get_latest(self, pre_ok=False, major=None):
        """
        Gets latest release satisfying "prereleases are OK" or major/branch constraints
        Strives to fetch formal API release if it exists, because it has useful information
        like assets.
        """
        # data of selected tag, always contains ['version', 'tag_name', 'tag_date', 'type'] will
        # be returned
        ret = None

        # always fetch /releases as this will allow us to quickly look up if a tag is marked as
        # pre-release

        # then always get *all* tags through pagination

        # if pre not ok, filter out tags to check

        # if major, filter out tags to check for major

        # we need to check all tags commit dates simply because the most recent wins
        # we don't check tags which:
        # * marked pre-release in releases endpoints
        # * has a beta-like, non-version tag name

        feed_entries = self.get_releases_feed_entries()
        if feed_entries:
            for tag in feed_entries:
                # https://github.com/apache/incubator-pagespeed-ngx/releases/tag/v1.13.35.2-stable
                tag_name = tag['link'].split('/')[-1]
                log.info('Checking tag {}'.format(tag_name))
                version = self.sanitize_version(tag_name, pre_ok, major)
                if not version:
                    log.info('We did not find a valid version in {} tag'.format(tag_name))
                    continue
                if ret and ret['version'] >= version:
                    log.info(
                        'Tag {} does not contain newer version than we already found'.format(tag_name)
                    )
                    continue
                tag_date = parser.parse(tag['updated'])
                if ret and tag_date + timedelta(days=365) < ret['tag_date']:
                    log.info('The version {} is newer, but is too old!'.format(version))
                    break
                # we always want to return formal release if it exists, cause it has useful data
                # grab formal release via APi to check for pre-release mark
                formal_release = self.get_formal_release_for_tag(tag_name)
                if formal_release:
                    # use full release info
                    ret = self.set_matching_formal_release(ret, formal_release, version, pre_ok)
                else:
                    if self.having_asset:
                        continue
                    log.info('No formal release for tag {}'.format(tag_name))
                    tag['tag_name'] = tag_name
                    tag['tag_date'] = tag_date
                    tag['version'] = version
                    tag['type'] = 'feed'
                    # remove keys which are non-jsonable
                    # TODO use those (pop returns them)
                    tag.pop('updated_parsed', None)
                    tag.pop('published_parsed', None)
                    ret = tag
                    log.info("Selected version as current selection: {}.".format(version))

        # we are good with release from feeds only without looking at the API
        # simply because feeds list stuff in order of recency
        # however, still use /tags unless releases.atom has data within a year
        if ret and ret['tag_date'].replace(tzinfo=None) > (datetime.utcnow() - timedelta(days=365)):
            return ret

        # only if we did not find desired stuff through feeds, we switch to using API :)
        # this may be required in cases
        # releases.atom has limited/no tags (#63), and all those are beta / invalid / non-versions
        # likewise, we want an older branch (major), which is not there in releases.atom
        # due to limited nature of data inside it

        # releases/latest fetches only non-prerelease, non-draft, so it
        # should not be used for hunting down pre-releases assets
        if not pre_ok:
            # https://stackoverflow.com/questions/28060116/which-is-more-reliable-for-github-api-conditional-requests-etag-or-last-modifie/57309763?noredirect=1#comment101114702_57309763
            # ideally we disable ETag validation for this endpoint completely
            r = self.repo_query('/releases/latest')
            if r.status_code == 200:
                formal_release = r.json()
                version = self.sanitize_version(formal_release['tag_name'], pre_ok, major)
                if version:
                    ret = self.set_matching_formal_release(ret, formal_release, version,
                                                           pre_ok, 'releases-latest')
        else:
            r = self.repo_query('/releases')
            if r.status_code == 200:
                for release in r.json():
                    tag_name = release['tag_name']
                    version = self.sanitize_version(tag_name, pre_ok, major)
                    if not version:
                        continue
                    if not ret or version > ret['version']:
                        ret = self.set_matching_formal_release(ret, release, version, pre_ok)

        if self.having_asset:
            # only formal releases which we enumerated above already, have assets
            # so there is no point looking in the tags/graphql below
            # return whatever we got
            return ret

        # formal release may not exist at all, or be "late/old" in case
        # actual release is only a simple tag so let's try /tags
        if self.api_token:
            # GraphQL requires auth
            ret = self.find_in_tags_via_graphql(ret, pre_ok, major)
        else:
            ret = self.find_in_tags(ret, pre_ok, major)

        return ret

    def set_matching_formal_release(self, ret, formal_release, version, pre_ok,
                                    data_type='release'):
        """Set current release selection to this formal release if matching conditions."""
        if not pre_ok and formal_release['prerelease']:
            log.info(
                "Found formal release for this tag which is unwanted "
                "pre-release: {}.".format(version))
            return ret
        if self.having_asset:
            if 'assets' not in formal_release or not formal_release['assets']:
                log.info('Skipping this release due to no assets.')
                return ret
            if self.having_asset is not True:
                regex_matching = False
                search = self.having_asset
                if search.startswith('~'):
                    search = r'{}'.format(search.lstrip('~'))
                    regex_matching = True
                found_asset = False
                for asset in formal_release['assets']:
                    if asset_matches(asset, search, regex_matching=regex_matching):
                        found_asset = True
                        break
                if not found_asset:
                    log.info('Desired asset not found in the release.')
                    return ret
        # formal_release['tag_name'] = tag_name
        formal_release['tag_date'] = parser.parse(formal_release['published_at'])
        formal_release['version'] = version
        formal_release['type'] = data_type
        log.info("Selected version as current selection: {}.".format(formal_release['version']))
        return formal_release

    def try_get_official(self, repo):
        """Check existence of repo/repo

        Returns:
            str: updated repo
        """
        official_repo = "{repo}/{repo}".format(repo=repo)
        log.info('Checking existence of {}'.format(official_repo))
        r = self.get('https://{}/{}/releases.atom'.format(self.hostname, official_repo))
        # API requests are varied by cookie, we don't want serializer for cache fail because of that
        self.cookies.clear()
        if r.status_code == 200:
            self.feed_contents[official_repo] = r.text
            return official_repo
        return None



import logging
from io import BytesIO
import os
import platform
import re
import sys
import tarfile
import requests
import tqdm

log = logging.getLogger(__name__)


class ApiCredentialsError(Exception):
    """raised when there's an API error related to credentials"""


class BadProjectError(Exception):
    """raised when no such project exists"""


# matches os.name to known extensions that are meant *mostly* to run on it, and not other os.name-s
os_extensions = {
    'nt': ('.exe', '.msi', '.msi.asc', '.msi.sha256'),
    'posix': ('.tgz', '.tar.gz')
}

extension_distros = {
    'deb': ['ubuntu', 'debian'],
    'rpm': ['rhel', 'centos', 'fedora', 'amazon', 'cloudlinux'],
    'apk': ['alpine'],
    'darwin': ['dmg']
}

# matches *start* of sys.platform value to words in asset name
platform_markers = {
    'win': ['windows', 'win'],
    'linux': ['linux'],
    'darwin': ['osx', 'darwin'],
    'freebsd': ['freebsd', 'netbsd', 'openbsd']
}

# this is all too simple for now
non_amd64_markers = ['i386', 'i686', 'arm', 'arm64', '386', 'ppc64', 'armv7', 'armv7l',
                     'mips64', 'ppc64', 'mips64le', 'ppc64le', 'aarch64']


def asset_does_not_belong_to_machine(asset):
    """
    Checks whether a given asset name is likely unusable on this machine
    An asset belongs to machine as long as this returns False
    :param asset:
    :type asset: str
    :return:
    :rtype:
    """
    # replace underscore with dash so that our shiny word boundary regexes won't break
    asset = asset.replace('_', '-')
    # bail if asset's extension "belongs" to other OS-es (simple)
    for os_name, ext in os_extensions.items():
        if os.name != os_name and asset.endswith(ext):
            return True
    for pf, pf_words in platform_markers.items():
        if not sys.platform.startswith(pf):
            for pfWord in pf_words:
                r = re.compile(r'\b{}(\d+)?\b'.format(pfWord), flags=re.IGNORECASE)
                matches = r.search(asset)
                if matches:
                    return True
    if sys.platform == 'linux':
        import distro
        # Weeding out non-matching Linux distros
        for ext, ext_distros in extension_distros.items():
            if asset.endswith("." + ext) and distro.id() not in ext_distros:
                return True
    # weed out non-64 bit stuff from x86_64 bit OS
    # caution: may be false positive with 32 bit Python on 64 bit OS
    if platform.machine() in ['x86_64', 'AMD64']:
        for non_amd64_word in non_amd64_markers:
            r = re.compile(r'\b{}\b'.format(non_amd64_word), flags=re.IGNORECASE)
            matches = r.search(asset)
            if matches:
                return True
            r = re.compile(r'\barm\d+\b', flags=re.IGNORECASE)
            matches = r.search(asset)
            if matches:
                return True
    return False


# monkey patching older requests library's response class so it can use context manager
# https://github.com/psf/requests/issues/4136
def requests_response_patched_enter(self):
    return self


def requests_response_patched_exit(self, *args):
    self.close()


if not hasattr(requests.Response, '__exit__'):
    requests.Response.__enter__ = requests_response_patched_enter
    requests.Response.__exit__ = requests_response_patched_exit


def download_file(url, local_filename=None):
    """Download a URL to the given filename.

    Args:
        url (str): URL to download from
        local_filename (str): Destination filename

    Returns:
        str: Destination filename, on success

    """
    if local_filename is None:
        local_filename = url.split('/')[-1]
    try:
        # NOTE the stream=True parameter below
        with requests.get(url, stream=True) as r:
            r.raise_for_status()
            # content-length may be empty, default to 0
            file_size = int(r.headers.get('Content-Length', 0))
            bar_size = 1024
            # fetch 8 KB at a time
            chunk_size = 8192
            # how many bars are there in a chunk?
            chunk_bar_size = chunk_size / bar_size
            # bars are by KB
            num_bars = int(file_size / bar_size)

            pbar = tqdm.tqdm(
                disable=None,  # disable on non-TTY
                total=num_bars,
                unit='KB',
                desc='Downloading {}'.format(local_filename),
                leave=True  # progressbar stays
            )
            with open(local_filename, 'wb') as f:
                for chunk in r.iter_content(chunk_size=chunk_size):
                    if chunk:  # filter out keep-alive new chunks
                        f.write(chunk)
                        # we fetch 8 KB, so we update progress by +8x
                        pbar.update(chunk_bar_size)
            pbar.set_description('Downloaded {}'.format(local_filename))
            pbar.close()
    except KeyboardInterrupt:
        os.remove(local_filename)
        log.warning('Cancelled')
        os._exit(1)
    return local_filename


def extract_file(url):
    """Extract an archive while stripping the top level dir."""
    smart_members = []
    response = requests.get(url, stream=True)
    # TODO make real stream processing of tar.gz
    with tarfile.open(fileobj=BytesIO(response.raw.read()), mode="r:gz") as tar_file:
        all_members = tar_file.getmembers()
        if not all_members:
            log.critical('No or not an archive')
        root_dir = all_members[0].path
        root_dir_with_slash_len = len(root_dir) + 1
        for member in tar_file.getmembers():
            if member.path.startswith(root_dir + "/"):
                member.path = member.path[root_dir_with_slash_len:]
                smart_members.append(member)
        tar_file.extractall(members=smart_members)

def rpm_installed_version(name):
    """Get the installed version of a package with the given name.

    Args:
        name (str): Package name

    Returns:
        string: Version of the installed packaged, or None
    """
    try:
        import rpm
    except ImportError:
        return False
    ts = rpm.TransactionSet()
    mi = ts.dbMatch('name', name)
    if mi:
        for h in mi:
            return h['version']
    return None



from .ProjectHolder import ProjectHolder
import logging
import yaml
from six.moves.urllib.parse import urlparse

log = logging.getLogger(__name__)


class HelmChartRepoSession(ProjectHolder):

    def __init__(self, repo, hostname=None):
        super(HelmChartRepoSession, self).__init__()
        if not repo.endswith('Chart.yaml'):
            repo = repo.rstrip('/') + '/Chart.yaml'
        log.info('Helm Chart.yml: {}'.format(repo))
        self.url = repo

    def get_latest(self, pre_ok=False, major=None):
        # https://github.com/bitnami/charts/blob/master/bitnami/aspnet-core/Chart.yaml
        # https://raw.githubusercontent.com/bitnami/charts/master/bitnami/aspnet-core/Chart.yaml
        url = self.url
        host = urlparse(url).hostname
        if host in ['github.com']:
            url = url.replace('github.com', 'raw.githubusercontent.com').replace('/blob/', '/')
        r = self.get(url)
        chart_data = yaml.safe_load(r.text)
        return {
            'tag_name': None,
            'tag_date': None,
            'version': self.sanitize_version(chart_data['version'], pre_ok, major),
            'type': 'helm'
        }



import re

from packaging.version import Version as PackagingVersion, InvalidVersion


class Version(PackagingVersion):
    """
    This class abstracts handling of a project's versions. It implements the
    scheme defined in PEP 440. A `Version` instance is comparison
    aware and can be compared and sorted using the standard Python interfaces.

    This class is descendant from Version found in packaging.version,
    and implements some additional, "AI"-like normalization during instantiation.

    Args:
        version (str): The string representation of a version which will be
                      parsed and normalized before use.
    Raises:
        InvalidVersion: If the ``version`` does not conform to PEP 440 in
                        any way then this exception will be raised.
    """

    def fix_letter_post_release(self, match):
        self.fixed_letter_post_release = True
        return match.group(1) + '.post' + str(ord(match.group(2)))

    def __init__(self, version, char_fix_required=False):
        # type: (str) -> None
        self.fixed_letter_post_release = False

        # 4.27-chaos-preview-3 -> 4.27-chaos-pre3
        version = re.sub('-preview-(\\d+)', '-pre\\1', version, 1)
        # 5.0.0-early-access-2 -> 5.0.0-alpha2
        version = re.sub('-early-access-(\\d+)', '-alpha\\1', version, 1)

        # many times they would tag foo-1.2.3 which would parse to LegacyVersion
        # we can avoid this, by reassigning to what comes after the dash:
        parts = version.split('-')

        # TODO test v5.12-rc1-dontuse -> v5.12.rc1
        # go through parts separated by dot, detect beta level, and weed out numberless info:
        parts_n = []
        for part in parts:
            # help devel releases to be correctly identified
            # https://www.python.org/dev/peps/pep-0440/#developmental-releases
            if part in ['devel', 'test', 'dev']:
                part = 'dev0'
            else:
                # help post (patch) releases to be correctly identified (e.g. Magento 2.3.4-p2)
                # p12 => post12
                part = re.sub('^p(\\d+)$', 'post\\1', part, 1)
            if not part.isalpha():
                parts_n.append(part)

        if not parts_n:
            raise InvalidVersion("Invalid version: '{0}'".format(version))
        # remove *any* non-digits which appear at the beginning of the version string
        # e.g. Rhino1_7_13_Release does not even bother to put a delimiter...
        # such string at the beginning typically do not convey stability level
        # so we are fine to remove them (unlike the ones in the tail)
        parts_n[0] = re.sub('^[^0-9]+', '', parts_n[0], 1)

        # go back to full string parse out
        version = ".".join(parts_n)

        if char_fix_required:
            version = re.sub('(\\d)([a-z])$', self.fix_letter_post_release, version, 1)
        # release-3_0_2 is often seen on Mercurial holders
        # note that above code removes "release-" already so we are left with "3_0_2"
        if re.search(r'^(?:\d+_)+(?:\d+)', version):
            version = version.replace('_', '.')
        # finally, split by dot "delimiter", see if there are common words which are definitely
        # removable
        parts = version.split('.')
        version = []
        for p in parts:
            if p.lower() in ['release']:
                continue
            version.append(p)
        version = '.'.join(version)
        super(Version, self).__init__(version)

    @property
    def epoch(self):
        """
        An integer giving the version epoch of this Version instance
        """
        # type: () -> int
        _epoch = self._version.epoch  # type: int
        return _epoch

    @property
    def release(self):
        """
        A tuple of integers giving the components of the release segment
        of this Version instance; that is, the 1.2.3 part of the version
        number, including trailing zeroes but not including the epoch or
        any prerelease/development/postrelease suffixes
        """
        _release = self._version.release
        return _release

    @property
    def pre(self):
        _pre = self._version.pre
        return _pre

    @property
    def post(self):
        return self._version.post[1] if self._version.post else None

    @property
    def dev(self):
        return self._version.dev[1] if self._version.dev else None

    @property
    def local(self):
        if self._version.local:
            return ".".join(str(x) for x in self._version.local)
        return None

    def __str__(self):
        # type: () -> str
        parts = []

        # Epoch
        if self.epoch != 0:
            parts.append("{0}!".format(self.epoch))

        # Release segment
        parts.append(".".join(str(x) for x in self.release))

        # Pre-release
        if self.pre is not None:
            parts.append("".join(str(x) for x in self.pre))

        # Post-release
        if self.post is not None:
            if self.fixed_letter_post_release:
                parts.append("{0}".format(chr(self.post)))
            else:
                parts.append(".post{0}".format(self.post))

        # Development release
        if self.dev is not None:
            parts.append(".dev{0}".format(self.dev))

        # Local version segment
        if self.local is not None:
            parts.append("+{0}".format(self.local))

        return "".join(parts)



import datetime

from .ProjectHolder import ProjectHolder
import logging


log = logging.getLogger(__name__)


class SystemRepoSession(ProjectHolder):

    def __init__(self, repo, hostname=None):
        super(SystemRepoSession, self).__init__()
        self.set_repo(repo)

    def dnf_get_available_version(self, pre_ok, major):
        ret = None
        import dnf
        with dnf.Base() as base:
            RELEASEVER = dnf.rpm.detect_releasever(base.conf.installroot)
            base.conf.substitutions['releasever'] = RELEASEVER
            # Repositories are needed if we want to install anything.
            base.read_all_repos()
            # A sack is needed for querying.
            base.fill_sack()
            # A query matches all packages in sack
            q = base.sack.query()
            # https://dnf.readthedocs.io/en/latest/use_cases.html#id3
            # https: // dnf.readthedocs.io / en / latest / api_queries.html  # dnf.query.Query
            # Derived query matches only available packages
            a = q.available()
            a = a.filter(name=self.repo)
            for pkg in a:  # a only gets evaluated here
                version = self.sanitize_version(pkg.version, pre_ok, major)
                if not ret or ret['version'] < version:
                    ret = {
                        'version': version,
                        'tag_name': pkg.evr,
                        'tag_date': datetime.datetime.fromtimestamp(pkg.buildtime)
                    }
        return ret

    def yum_get_available_version(self, pre_ok, major):
        ret = None
        import yum
        yumLoggers = ['yum.filelogging.RPMInstallCallback', 'yum.verbose.Repos',
                      'yum.verbose.plugin',
                      'yum.Depsolve', 'yum.verbose', 'yum.plugin', 'yum.Repos', 'yum',
                      'yum.verbose.YumBase',
                      'yum.filelogging', 'yum.verbose.YumPlugins', 'yum.RepoStorage', 'yum.YumBase',
                      'yum.filelogging.YumBase', 'yum.verbose.Depsolve']
        for logger in yumLoggers:
            logging.getLogger(logger).setLevel(logging.CRITICAL)
        yb = yum.YumBase()
        yb.preconf.debuglevel = 0
        yb.preconf.errorlevel = 0
        yb.setCacheDir()
        pkgs = yb.pkgSack.returnNewestByNameArch(patterns=[self.repo])
        for pkg in pkgs:
            version = self.sanitize_version(pkg.vr, pre_ok, major)
            if not ret or ret['version'] < version:
                ret = {
                    'version': version,
                    'tag_name': pkg.vr,
                    # 'tag_date': ?
                }
        return ret

    def apt_get_available_version(self, pre_ok, major):
        ret = None
        import apt
        cache = apt.cache.Cache()
        cache.update()
        cache.open()
        if self.repo in cache:
            pkg = cache[self.repo]
            for pkg_ver in pkg.versions:
                version = pkg_ver.version.split('-')[0]
                version = self.sanitize_version(version, pre_ok, major)
                if not ret or ret['version'] < version:
                    ret = {
                        'version': version,
                        'tag_name': pkg_ver.version,
                        # 'tag_date': ?
                    }
        return ret

    def get_latest(self, pre_ok=False, major=None):
        try:
            return self.dnf_get_available_version(pre_ok, major)
        except ImportError:
            pass
        try:
            return self.yum_get_available_version(pre_ok, major)
        except ImportError:
            pass
        try:
            return self.apt_get_available_version(pre_ok, major)
        except ImportError:
            pass
        return None



import datetime

import feedparser

from .ProjectHolder import ProjectHolder


class SourceForgeRepoSession(ProjectHolder):
    """SourceForce project holder."""

    REPO_URL_PROJECT_COMPONENTS = 1
    DEFAULT_HOSTNAME = 'sourceforge.net'
    # For project URLs, e.g. https://sourceforge.net/projects/keepass/
    # a URI does not start with a repo name, skip '/projects/'
    REPO_URL_PROJECT_OFFSET = 1

    def __init__(self, repo, hostname):
        super(SourceForgeRepoSession, self).__init__()
        self.hostname = hostname
        self.set_repo(repo)

    def get_latest(self, pre_ok=False, major=None):
        ret = None
        # to leverage cachecontrol, we fetch the feed using requests as usual
        # then feed the feed to feedparser as a raw string
        # e.g. https://sourceforge.net/projects/keepass/rss?path=/
        # TODO this could be better. Now it is actually checking versions in topmost files
        r = self.get('https://{}/projects/{}/rss?path=/'.format(self.hostname, self.repo))
        feed = feedparser.parse(r.text)
        for tag in feed.entries:
            tag_name = tag['title']
            version = self.sanitize_version(tag_name, pre_ok, major)
            if not version:
                continue
            if not ret or version > ret['version']:
                ret = tag
                tag['tag_name'] = tag['title']
                tag['version'] = version
                # converting from struct
                tag['tag_date'] = datetime.datetime(*tag['published_parsed'][:6])
        return ret



import datetime
import logging

import feedparser

from .ProjectHolder import ProjectHolder

log = logging.getLogger(__name__)


class FeedRepoSession(ProjectHolder):
    KNOWN_REPOS_BY_NAME = {
        'filezilla': {
            'repo': 'filezilla',
            'hostname': 'filezilla-project.org',
            'only': 'FileZilla Client'
        }
    }

    # https://alex.miller.im/posts/python-3-feedfinder-rss-detection-from-url/
    def find_feed(self, site):
        from six.moves.urllib.parse import urlparse
        from bs4 import BeautifulSoup as bs4
        raw = self.get(site).text
        result = []
        possible_feeds = []
        html = bs4(raw, "html.parser")
        self.home_soup = html
        feed_urls = html.findAll("link", rel="alternate")

        for f in feed_urls:
            t = f.get("type", None)
            if not t:
                continue
            if "rss" in t or "xml" in t:
                href = f.get("href", None)
                if href:
                    possible_feeds.append(href)
        parsed_url = urlparse(site)
        base = parsed_url.scheme + "://" + parsed_url.hostname
        a_tags = html.findAll("a")
        for a in a_tags:
            href = a.get("href", None)
            if not href:
                continue
            if "xml" in href or "rss" in href or "feed" in href:
                possible_feeds.append(base + '/' + href.lstrip('/'))
        for url in list(set(possible_feeds)):
            f = feedparser.parse(url)
            if len(f.entries) > 0 and url not in result:
                result.append(url)
        return result

    def __init__(self, repo, hostname):
        super(FeedRepoSession, self).__init__()
        self.home_soup = None
        feeds = self.find_feed('https://' + hostname + '/')
        if not feeds:
            return
        self.hostname = hostname
        self.set_repo(repo)
        log.info('Using feed URL: {}'.format(feeds[0]))
        self.feed_url = feeds[0]

    def get_latest(self, pre_ok=False, major=None):
        ret = None
        # to leverage cachecontrol, we fetch the feed using requests as usual
        # then feed the feed to feedparser as a raw string
        # e.g. https://hg.nginx.org/nginx/atom-tags
        # https://pythonhosted.org/feedparser/common-atom-elements.html
        r = self.get(self.feed_url)
        feed = feedparser.parse(r.text)
        for tag in feed.entries:
            tag_name = tag['title']
            version = self.sanitize_version(tag_name, pre_ok, major)
            if not version:
                continue
            if not ret or version > ret['version']:
                ret = tag
                tag['tag_name'] = tag['title']
                tag['version'] = version
                if 'published_parsed' in tag:
                    # converting from struct
                    tag['tag_date'] = datetime.datetime(*tag['published_parsed'][:6])
                elif 'updated_parsed' in tag:
                    tag['tag_date'] = datetime.datetime(*tag['updated_parsed'][:6])
        return ret



# Maps SPDX ID from https://spdx.org/licenses/
# to corresponding short name license as recommended by Fedora
# https://fedoraproject.org/wiki/Licensing:Main?rd=Licensing#Good_Licenses
rpmspec_licenses = {
   'Apache-2.0': 'APSL 2.0',
   'BSD-2-Clause': 'BSD',
   'MIT': 'MIT',
}



import logging

from dateutil import parser

from .ProjectHolder import ProjectHolder
from .utils import BadProjectError

log = logging.getLogger(__name__)


class PypiRepoSession(ProjectHolder):
    """A class to represent a Pypi project holder."""
    DEFAULT_HOSTNAME = 'pypi.org'
    REPO_URL_PROJECT_COMPONENTS = 1
    # For project URLs, e.g. https://pypi.org/project/lastversion/
    # a URI does not start with a repo name, skip '/project/'
    REPO_URL_PROJECT_OFFSET = 1

    def get_project(self):
        """Get project JSON data."""
        project = None
        url = 'https://{}/pypi/{}/json'.format(self.hostname, self.repo)
        log.info('Requesting {}'.format(url))
        r = self.get(url)
        if r.status_code == 200:
            project = r.json()
        return project

    def __init__(self, repo, hostname=None):
        super(PypiRepoSession, self).__init__()
        if hostname:
            self.hostname = hostname
        else:
            self.hostname = PypiRepoSession.DEFAULT_HOSTNAME
        self.set_repo(repo)
        self.project = self.get_project()
        if hostname and not self.project:
            raise BadProjectError(
                'The project {} does not exist on Pypi'.format(
                    repo
                )
            )

    def release_download_url(self, release, shorter=False):
        """Get release download URL."""
        for f in release['files']:
            if f['packagetype'] == 'sdist':
                return f['url']
        return None

    def get_latest(self, pre_ok=False, major=None):
        """Get latest project release."""
        ret = self.project
        # we are in "enriching" project dict with desired version information
        # and return None if there's no matching version
        from .Version import Version
        if not major:
            latest_ver = self.project['info']['version']
            v = Version(latest_ver)
            ret['version'] = v
            # there are no tags, we just put version string there
            ret['tag_name'] = latest_ver
        else:
            for release_ver in self.project['releases']:
                version = self.sanitize_version(release_ver, pre_ok, major)
                if not version:
                    continue
                if 'version' not in ret or version > ret['version']:
                    ret['tag_name'] = release_ver
                    ret['version'] = version
        if 'tag_name' in ret:
            # consider tag_date as upload time of the selected release first file
            ret['files'] = self.project['releases'][ret['tag_name']]
            ret['tag_date'] = parser.parse(ret['files'][0]['upload_time'])
            return ret
        return None

    @staticmethod
    def make_canonical_link(repo):
        return 'https://{}/project/{}/'.format(PypiRepoSession.DEFAULT_HOSTNAME, repo)

    def get_canonical_link(self):
        return 'https://{}/project/{}/'.format(self.hostname, self.repo)



import logging

from .ProjectHolder import ProjectHolder
from .utils import BadProjectError

log = logging.getLogger(__name__)


class WordPressPluginRepoSession(ProjectHolder):
    """A class to represent a WordPress plugin  project holder."""
    DEFAULT_HOSTNAME = 'wordpress.org'
    REPO_URL_PROJECT_COMPONENTS = 1
    # For project URLs, e.g. https://wordpress.org/plugins/opcache-reset/
    # a URI does not start with a repo name, skip '/plugins/'
    REPO_URL_PROJECT_OFFSET = 1

    def get_project(self):
        """Get project JSON data."""
        project = None
        url = 'https://api.{}/plugins/info/1.0/{}.json'.format(self.hostname, self.repo)
        log.info('Requesting {}'.format(url))
        r = self.get(url)
        if r.status_code == 200:
            project = r.json()
        return project

    def __init__(self, repo, hostname=None):
        super(WordPressPluginRepoSession, self).__init__()
        if hostname:
            self.hostname = hostname
        else:
            self.hostname = WordPressPluginRepoSession.DEFAULT_HOSTNAME
        self.set_repo(repo)
        self.project = self.get_project()
        if hostname and not self.project:
            raise BadProjectError(
                'The project {} does not exist in WordPress plugin directory'.format(
                    repo
                )
            )

    def release_download_url(self, release, shorter=False):
        """Get release download URL."""
        return 'https://downloads.wordpress.org/plugin/{}.{}.zip'.format(
            self.repo, release['version'])

    def get_latest(self, pre_ok=False, major=None):
        """Get latest release for this project."""
        ret = {}
        # we are in "enriching" project dict with desired version information
        # and return None if there's no matching version
        from .Version import Version
        if not major:
            latest_ver = self.project['version']
            v = Version(latest_ver)
            ret['version'] = v
            # there are no tags, we just put version string there
            ret['tag_name'] = latest_ver
        else:
            for release_ver in self.project['versions']:
                version = self.sanitize_version(release_ver, pre_ok, major)
                if not version:
                    continue
                if 'version' not in ret or version > ret['version']:
                    ret['tag_name'] = release_ver
                    ret['version'] = version
                    log.info('Set current selection to {}'.format(version))
                else:
                    log.info('Not set {}'.format(str(version)))
        if 'tag_name' in ret:
            self.project.update(ret)
            return self.project
        return None

    @staticmethod
    def make_canonical_link(repo):
        """Make canonical link from repo."""
        return 'https://{}/plugins/{}/'.format(WordPressPluginRepoSession.DEFAULT_HOSTNAME, repo)

    def get_canonical_link(self):
        """Get canonical link from repo."""
        return 'https://{}/plugins/{}/'.format(self.hostname, self.repo)



# -*- coding: utf-8 -*-
# License: BSD, see LICENSE for more details.
"""
This is the main module of lastversion package.
To use it, import it and invoke any function documented here. For example:

```python
from lastversion import lastversion
lastversion.has_update(repo='mautic/mautic', current_version='1.2.3')
```
"""

import argparse
import json
import logging
import os
import re
import sys

import yaml
from os.path import expanduser

from packaging.version import InvalidVersion

from .__about__ import __self__
from .GitHubRepoSession import TOKEN_PRO_TIP
from .HolderFactory import HolderFactory
from .ProjectHolder import ProjectHolder
from .Version import Version
from .argparse_version import VersionAction
from .utils import download_file, extract_file, rpm_installed_version, ApiCredentialsError, \
    BadProjectError
from .spdx_id_to_rpmspec import rpmspec_licenses
from six.moves.urllib.parse import urlparse

log = logging.getLogger(__name__)


def latest(repo, output_format='version', pre_ok=False, assets_filter=None,
           short_urls=False, major=None, only=None, at=None,
           having_asset=None, exclude=None):
    """Find latest release version for a project.

    Args:
        major (str): Only consider versions which are "descendants" of this major version string
        short_urls (bool): Whether we should try to return shorter URLs for release data
        assets_filter (str): Regular expression for filtering assets for the latest release
        only (str): Only consider tags with this text. Useful for repos with multiple projects.
                    The argument supports negation and regular expressions. To indicate a regex,
                    start it with tilde sign, to negate the expression, start it with exclamtion
                    point, e.g. r"!~\w" will make it consider only tags without word characters.
        repo (str): Repository specifier in any form.
        output_format (str): Affects return format. Possible values `version`, `json`, `dict`,
                             `assets`, `source`, `tag`.
        pre_ok (bool): Specifies whether pre-releases can be accepted as newer version.
        at (str): Specifies repo hosting more precisely, only useful if repo argument was
                  specified as one word.
        having_asset (Union[str, bool]): Only consider releases with the given asset.
                                         Pass `True` for any asset

    Returns:
        Version: Newer version object, if found and `output_format` is `version`.
    Returns:
        str: Single string containing tag, if found and `output_format` is `tag`

    """
    repo_data = {}

    if repo.endswith('.yml') and not repo.startswith(('http://', 'https://')):
        with open(repo) as fpi:
            repo_data = yaml.safe_load(fpi)
            if 'repo' in repo_data:
                if 'nginx-extras' in repo:
                    repo_data['module_of'] = 'nginx'
                name = os.path.splitext(os.path.basename(repo))[0]
                if 'module_of' in repo_data:
                    name = '{}-module-{}'.format(repo_data['module_of'], name)
                repo = repo_data['repo']
                repo_data['name'] = name

    if repo.startswith(('http://', 'https://')) and repo.endswith('Chart.yaml'):
        at = 'helm_chart'

    if repo.endswith('.spec'):
        # repo is specified inside the .spec file
        # github repo is resolved via %{upstream_github} + %{name}/%{upstream_name}
        # no upstream_github global means that the spec was not prepared for lastversion
        # optional: use of spec_tag macros if the source is from GitHub. in edge cases we check
        # new version via GitHub, but prepared sources are elsewhere
        with open(repo) as f:
            name = None
            upstream_github = None
            upstream_name = None
            current_version = None
            spec_repo = None
            spec_url = None
            for l in f.readlines():
                if l.startswith('%global lastversion_repo'):
                    spec_repo = l.split(' ')[2].strip()
                elif l.startswith('%global upstream_github'):
                    upstream_github = l.split(' ')[2].strip()
                elif l.startswith('%global upstream_name'):
                    upstream_name = l.split(' ')[2].strip()
                elif l.startswith('Name:'):
                    name = l.split('Name:')[1].strip()
                elif l.startswith('URL:'):
                    spec_url = l.split('URL:')[1].strip()
                elif l.startswith('%global upstream_version '):
                    current_version = l.split(' ')[2].strip()
                    # influences %spec_tag to use %upstream_version instead of %version
                    repo_data['module_of'] = True
                elif l.startswith('Version:') and not current_version:
                    current_version = l.split('Version:')[1].strip()
            if spec_url:
                spec_host = urlparse(spec_url).hostname
                if spec_host in ['github.com'] and not upstream_github and not spec_repo:
                    log.warning('Neither %upstream_github nor %lastversion_repo macros were found. '
                                'Please prepare your spec file using instructions: '
                                'https://lastversion.getpagespeed.com/spec-preparing.html')
            if not current_version:
                log.critical('Did not find neither Version: nor %upstream_version in the spec file')
                sys.exit(1)
            try:
                if current_version != 'x':
                    repo_data['current_version'] = Version(current_version)
            except InvalidVersion:
                log.critical('Failed to parse current version in {}. Tried {}'.format(
                    repo, current_version))
                sys.exit(1)
            if upstream_name:
                repo_data['name'] = upstream_name
                repo_data['spec_name'] = '%{upstream_name}'
            else:
                repo_data['name'] = name
                repo_data['spec_name'] = '%{name}'
            if upstream_github:
                repo = "{}/{}".format(upstream_github, repo_data['name'])
                log.info('Discovered GitHub repo {} from .spec file'.format(repo))
            elif spec_repo:
                repo = spec_repo
                log.info('Discovered explicit repo {} from .spec file'.format(repo))
            elif spec_url:
                repo = spec_url

    with HolderFactory.get_instance_for_repo(repo, at=at) as project:
        project.set_only(only)
        project.set_exclude(exclude)
        project.set_having_asset(having_asset)
        release = project.get_latest(pre_ok=pre_ok, major=major)

        # bail out, found nothing that looks like a release
        if not release:
            return None

        from_type = 'Located the latest release tag {} at: {}'.format(
            release['tag_name'], project.get_canonical_link()
        )
        if 'type' in release:
            from_type = '{} via {} mechanism'.format(from_type, release['type'])
        log.info(from_type)

        version = release['version']
        tag = release['tag_name']

        # return the release if we've reached far enough:
        if output_format == 'version':
            return version

        if output_format in ['json', 'dict']:
            if output_format == 'dict':
                release['version'] = version
            else:
                release['version'] = str(version)
                if 'tag_date' in release:
                    release['tag_date'] = str(release['tag_date'])
            release['v_prefix'] = tag.startswith("v")
            version_macro = 'upstream_version' if 'module_of' in repo_data else 'version'
            version_macro = '%{{{}}}'.format(version_macro)
            holder_i = {value: key for key, value in HolderFactory.HOLDERS.items()}
            release['source'] = holder_i[type(project)]
            release['spec_tag'] = tag.replace(
                str(version),
                version_macro
            )
            # spec_tag_no_prefix is the helpful macro which will allow us to know where tarball
            # extracts to (GitHub-specific)
            if release['spec_tag'].startswith('v{}'.format(version_macro)) or \
                    re.match(r'^v\d', release['spec_tag']):
                release['spec_tag_no_prefix'] = release['spec_tag'].lstrip('v')
            else:
                release['spec_tag_no_prefix'] = release['spec_tag']
            release['tag_name'] = tag
            if hasattr(project, 'repo_license'):
                release['license'] = project.repo_license(tag)
            if hasattr(project, 'repo_readme'):
                release['readme'] = project.repo_readme(tag)
            release.update(repo_data)
            try:
                release['assets'] = project.get_assets(release, short_urls, assets_filter)
            except NotImplementedError:
                pass
            release['from'] = project.get_canonical_link()

            spdx_id = release['license']['license']['spdx_id'] if 'license' in release else None
            rpmspec_licence = rpmspec_licenses[spdx_id] if spdx_id in rpmspec_licenses else None
            if rpmspec_licence:
                release['rpmspec_license'] = rpmspec_licence
            return release

        if output_format == 'assets':
            return project.get_assets(release, short_urls, assets_filter)

        if output_format == 'source':
            return project.release_download_url(release, short_urls)

        if output_format == 'tag':
            return tag

    return None


def has_update(repo, current_version, pre_ok=False, at=None):
    """Given an existing version for a repo, checks if there is an update.

    Args:
        repo (str): Repository specifier in any form.
        current_version (str): A version you want to check update for.
        pre_ok (bool): Specifies whether pre-releases can be accepted as newer version.
        at (str): Specifies repo hosting more precisely, only useful if repo argument was
                  specified as one word.

    Returns:
        Version: Newer version as object, if found. Otherwise, False

    """
    latest_version = latest(repo, output_format='version', pre_ok=pre_ok, at=at)
    if latest_version and latest_version > Version(current_version):
        return latest_version
    return False


def check_version(value):
    """Given a version string, raises argparse.ArgumentTypeError if it does not contain any version.
    In lastversion CLI app, this is used as argument parser helper for --newer-than (-gt) option.

    Args:
        value (str): Free-format string which is meant to contain user-supplied version

    Raises:
        argparse.ArgumentTypeError: Exception in case version was not found in the input string

    Returns:
        Version: Parsed version object

    """
    """
    Argument parser helper for --newer-than (-gt) option
    :param value:
    :type value:
    :return:
    :rtype:
    """
    try:
        # TODO use sanitize_version so that we can just pass tags as values
        # help devel releases to be correctly identified
        # https://www.python.org/dev/peps/pep-0440/#developmental-releases
        value = re.sub('-devel$', '.dev0', value, 1)
        # help post (patch) releases to be correctly identified (e.g. Magento 2.3.4-p2)
        value = re.sub('-p(\\d+)$', '.post\\1', value, 1)
        value = Version(value)
    except InvalidVersion:
        raise argparse.ArgumentTypeError("%s is an invalid version value" % value)
    return value


def parse_version(tag):
    """Parse version to Version object."""
    h = ProjectHolder()
    return h.sanitize_version(tag, pre_ok=True)


def get_rpm_packager():
    try:
        rpmmacros = expanduser("~") + "/.rpmmacros"
        with open(rpmmacros) as f:
            for ln in f.readlines():
                if ln.startswith('%packager'):
                    return ln.split('%packager')[1].strip()
    except IOError:
        log.warning("~/.rpmmacros does not exist. Changelog will not be generated")
    return None


def update_spec(repo, res, sem='minor'):
    print(res['version'])
    if 'current_version' not in res or res['current_version'] < res['version']:
        log.info('Updating spec {} with semantic {}'.format(repo, sem))
        if 'current_version' in res and len(res['version'].release) >= 3:
            current_major = res['current_version'].release[0]
            latest_major = res['version'].release[0]
            current_minor = res['current_version'].release[1]
            latest_minor = res['version'].release[1]
            if sem in ['minor', 'patch']:
                fail_fmt = 'Latest v{} fails semantic {} constraint against current v{}'
                if latest_major != current_major:
                    log.warning(
                        fail_fmt.format(res['version'], sem, res['current_version']))
                    sys.exit(4)
                if sem == 'patch' and latest_minor != current_minor:
                    log.warning(
                        fail_fmt.format(res['version'], sem, res['current_version']))
                    sys.exit(4)
    else:
        log.info('No newer version than already present in spec file')
        sys.exit(2)
    # update %lastversion_tag and %lastversion_dir, Version (?), Release
    out = []
    packager = get_rpm_packager()
    with open(repo) as f:
        for ln in f.readlines():
            if ln.startswith('%global lastversion_tag '):
                out.append('%global lastversion_tag {}'.format(res['spec_tag']))
            elif ln.startswith('%global lastversion_dir '):
                out.append('%global lastversion_dir {}-{}'.format(
                    res['spec_name'], res['spec_tag_no_prefix']))
            elif ln.startswith('%global upstream_version '):
                out.append('%global upstream_version {}'.format(res['version']))
            elif ln.startswith('Version:') and ('module_of' not in res or not res['module_of']):
                version_tag_regex = r'^Version:(\s+)(\S+)'
                m = re.match(version_tag_regex, ln)
                out.append('Version:' + m.group(1) + str(res['version']))
            elif ln.startswith('%changelog') and packager:
                from datetime import datetime
                now = datetime.utcnow()
                today = now.strftime('%a %b %d %Y')
                out.append(ln.rstrip())
                out.append('* {} {}'.format(today, packager))
                out.append('- upstream release v{}'.format(res['version']))
                out.append("\n")
            elif ln.startswith('Release:'):
                release_tag_regex = r'^Release:(\s+)(\S+)'
                m = re.match(release_tag_regex, ln)
                release = m.group(2)
                from string import digits
                release = release.lstrip(digits)
                out.append('Release:' + m.group(1) + '1' + release)
            else:
                out.append(ln.rstrip())
    with open(repo, "w") as f:
        f.write("\n".join(out))


def main():
    """The entrypoint to CLI app."""
    epilog = None
    if "GITHUB_API_TOKEN" not in os.environ and "GITHUB_TOKEN" not in os.environ:
        epilog = TOKEN_PRO_TIP
    parser = argparse.ArgumentParser(description='Find the latest software release.',
                                     epilog=epilog,
                                     prog='lastversion')
    parser.add_argument('action', nargs='?', default='get',
                        help='Special action to run, e.g. download, install, test')
    parser.add_argument('repo', metavar='<repo or URL>',
                        help='Repository in format owner/name or any URL that belongs to it')
    # affects what is considered last release
    parser.add_argument('--pre', dest='pre', action='store_true',
                        help='Include pre-releases in potential versions')
    parser.add_argument('--sem', dest='sem', choices=['major', 'minor', 'patch'],
                        help='Semantic version constraint against compared version')
    parser.add_argument('-v', '--verbose', action='count', default=0,
                        help='Will give you an idea of what is happening under the hood, '
                             '-vv to increase verbosity level')
    # no --download = False, --download filename.tar, --download = None
    parser.add_argument('-d', '--download', dest='download', nargs='?', default=False, const=None,
                        metavar='FILENAME', help='Download with custom filename')
    # how / which data of last release we want to present
    # assets will give download urls for assets if available and sources archive otherwise
    # sources will give download urls for sources always
    # json always includes "version", "tag_name" etc + whichever json data was
    # used to satisfy lastversion
    parser.add_argument('--format',
                        choices=['version', 'assets', 'source', 'json', 'tag'],
                        help='Output format')
    parser.add_argument('--assets', dest='assets', action='store_true',
                        help='Returns assets download URLs for last release')
    parser.add_argument('--source', dest='source', action='store_true',
                        help='Returns only source URL for last release')
    parser.add_argument('-gt', '--newer-than', type=check_version, metavar='VER',
                        help="Output only if last version is newer than given version")
    parser.add_argument('-b', '--major', '--branch', metavar='MAJOR',
                        help="Only consider releases of a specific major "
                             "version, e.g. 2.1.x")
    parser.add_argument('--only', metavar='REGEX', 
                        help="Only consider releases containing this text. "
                             "Useful for repos with multiple projects inside")
    parser.add_argument('--exclude', metavar='REGEX',
                        help="Only consider releases NOT containing this text. "
                             "Useful for repos with multiple projects inside")
    parser.add_argument('--filter', metavar='REGEX', help="Filters --assets result by a regular "
                                                          "expression")
    parser.add_argument('--having-asset', metavar='ASSET',
                        help="Only consider releases with this asset",
                        nargs='?', const=True)
    parser.add_argument('-su', '--shorter-urls', dest='shorter_urls', action='store_true',
                        help='A tiny bit shorter URLs produced')
    parser.add_argument('--at', dest='at',
                        help='If the repo argument is one word, specifies where to look up the '
                             'project. The default is via internal lookup or GitHub Search',
                        choices=HolderFactory.HOLDERS.keys())
    parser.add_argument('-y', '--assumeyes', dest='assumeyes', action='store_true',
                        help='Automatically answer yes for all questions')
    parser.add_argument('--version', action=VersionAction)
    parser.set_defaults(validate=True, verbose=False, format='version',
                        pre=False, assets=False, newer_than=False, filter=False,
                        shorter_urls=False, major=None, assumeyes=False, at=None,
                        having_asset=None)
    args = parser.parse_args()

    if args.repo == "self":
        args.repo = __self__

    # "expand" repo:1.2 as repo --branch 1.2
    if ':' in args.repo and \
            not (args.repo.startswith(('https://', 'http://')) and args.repo.count(':') == 1):
        # right split ':' once only to preserve it in protocol of URLs
        # https://github.com/repo/owner:2.1
        repo_args = args.repo.rsplit(':', 1)
        args.repo = repo_args[0]
        args.major = repo_args[1]

    # instead of using root logger, we use
    logger = logging.getLogger('lastversion')
    # create console handler and set level to debug
    ch = logging.StreamHandler()
    # create formatter
    fmt = '%(name)s - %(levelname)s - %(message)s' if args.verbose else '%(levelname)s: %(message)s'
    formatter = logging.Formatter(fmt)
    # add formatter to ch
    ch.setFormatter(formatter)
    # add ch to logger
    logger.addHandler(ch)

    if args.verbose:
        logger.setLevel(logging.DEBUG)
        log.info("Verbose {} level output.".format(args.verbose))
        if args.verbose >= 2:
            cachecontrol_logger = logging.getLogger('cachecontrol')
            cachecontrol_logger.removeHandler(logging.NullHandler())
            cachecontrol_logger.addHandler(ch)
            cachecontrol_logger.setLevel(logging.DEBUG)

    if args.assets:
        args.format = 'assets'

    if args.source:
        args.format = 'source'

    if args.filter:
        args.filter = re.compile(args.filter)

    if args.action == 'test':
        v = parse_version(args.repo)
        if not v:
            print('Failed to parse as a valid version')
            sys.exit(1)
        else:
            print("Parsed as: {}".format(v))
            print("Stable: {}".format(not v.is_prerelease))
            sys.exit(0)

    if args.action == 'install':
        # we can only install assets
        args.format = 'json'
        if args.having_asset is None:
            args.having_asset = r'~\.rpm$'
            try:
                import apt
                args.having_asset = r'~\.deb$'
            except ImportError:
                pass

    if args.repo.endswith('.spec'):
        args.action = 'update-spec'
        args.format = 'dict'

    if not args.sem:
        if args.action == 'update-spec':
            args.sem = 'minor'
        else:
            args.sem = 'any'
    # imply source download, unless --assets specified
    # --download is legacy flag to specify download action or name of desired download file
    # --download == None indicates download intent where filename is based on upstream
    if args.action == 'download' and args.download is False:
        args.download = None

    if args.download is not False:
        args.action = 'download'
        if args.format != 'assets':
            args.format = 'source'

    if args.action in ['extract', 'unzip'] and args.format != 'assets':
        args.format = 'source'

    if args.newer_than:
        base_compare = parse_version(args.repo)
        if base_compare:
            print(max([args.newer_than, base_compare]))
            sys.exit(2 if base_compare <= args.newer_than else 0)
     
    # other action are either getting release or doing something with release (extend get action)
    try:
        res = latest(args.repo, args.format, args.pre, args.filter,
                     args.shorter_urls, args.major, args.only, args.at,
                     having_asset=args.having_asset, exclude=args.exclude)
    except (ApiCredentialsError, BadProjectError) as error:
        sys.stderr.write(str(error) + os.linesep)
        if isinstance(error, ApiCredentialsError) and "GITHUB_API_TOKEN" not in os.environ and \
                "GITHUB_TOKEN" not in os.environ:
            sys.stderr.write(TOKEN_PRO_TIP + os.linesep)
        sys.exit(4)

    if res:
        if args.action == 'update-spec':
            return update_spec(args.repo, res, sem=args.sem)
        if args.action == 'download':
            # download command
            if args.format == 'source':
                # there is only one source, but we need an array
                res = [res]
            download_name = None
            # save with custom filename, if there's one file to download
            if len(res) == 1:
                download_name = args.download
            for url in res:
                log.info("Downloading {} ...".format(url))
                download_file(url, download_name)
            sys.exit(0)

        if args.action in ['unzip', 'extract']:
            # download command
            if args.format == 'source':
                # there is only one source, but we need an array
                res = [res]
            download_name = None
            # save with custom filename, if there's one file to download
            if len(res) == 1:
                download_name = args.download
            for url in res:
                log.info("Extracting {} ...".format(url))
                extract_file(url)
            sys.exit(0)

        if args.action == 'install':
            rpms = [asset for asset in res['assets'] if asset.endswith('.rpm')]
            if not rpms:
                log.error('No assets found to install')
                sys.exit(1)
            # prevents downloading large packages if we already have newest installed
            # consult RPM database  for current version
            installed_version = rpm_installed_version(args.repo)
            if installed_version is False:
                log.warning('Please install lastversion using YUM or DNF so it can check current '
                            'program version. This is helpful to prevent unnecessary downloads')
            if installed_version and Version(installed_version) >= Version(res['version']):
                log.warning('Newest version {} is already installed'.format(installed_version))
                sys.exit(0)
            # pass RPM URLs directly to package management program
            try:
                import subprocess
                params = ['yum', 'install']
                params.extend(rpms)
                if args.assumeyes:
                    params.append('-y')
                subprocess.call(params)
            except OSError:
                log.critical('Failed to launch package manager. Only YUM/DNF is supported!')
                sys.exit(1)
            # if the system has yum, then lastversion has to be installed from yum and
            # has access to system packages like yum python or dnf python API
            # if install_with_dnf(rpms) is False or install_with_yum(rpms) is False:
            #     log.error('Failed talking to either DNF or YUM for package install')
            #     sys.exit(1)
            sys.exit(0)

        # display version in various formats:
        if args.format == 'assets':
            print("\n".join(res))
        elif args.format == 'json':
            json.dump(res, sys.stdout)
        else:
            print(res)
            # special exit code "2" is useful for scripting to detect if no newer release exists
            if args.newer_than and res <= args.newer_than:
                sys.exit(2)
    else:
        # empty list returned to --assets, emit 3
        if args.format == 'assets' and res is not False:
            sys.exit(3)
        sys.stderr.write("No release was found" + os.linesep)
        sys.exit(1)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

import logging
import os
import re
from appdirs import user_cache_dir
from cachecontrol import CacheControlAdapter
from cachecontrol.caches.file_cache import FileCache
import requests
from packaging.version import InvalidVersion

from .Version import Version
from .__about__ import __version__
# this class basically corresponds to something (often a website) which holds
# projects (usually a bunch). often this is a github-like website, so we subclass session
# but this also maybe something special, which either way can be used as a source of version
# information for a project based on its URL or name (see LocalVersionSession)
# it is instantiated with a particular project in mind/set, but also has some methods for
# stuff like searching one
from .utils import asset_does_not_belong_to_machine

log = logging.getLogger(__name__)


class ProjectHolder(requests.Session):
    """Generic project holder class abstracts a web-accessible project storage."""

    """List of odd repos where last char is part of version not beta level."""
    LAST_CHAR_FIX_REQUIRED_ON = []

    # web accessible project holders may have single well-known domain usable by everyone
    # in case of GitHub, that is github.com, for Mercurial web gui - here isn't one, etc.
    DEFAULT_HOSTNAME = None
    DEFAULT_HOLDER = False
    KNOWN_REPO_URLS = {}
    KNOWN_REPOS_BY_NAME = {}
    # e.g. owner/project, but mercurial just /project together with hostname
    # adapter array should list how many elements make up "repo", e.g. for hg.nginx.com/repo it
    # is only one instead of 2
    # or a "format" specifier for matching
    REPO_URL_PROJECT_COMPONENTS = 2
    # if URI starts with project name, 0. Otherwise skip through this many URI dirs
    REPO_URL_PROJECT_OFFSET = 0
    RELEASE_URL_FORMAT = None
    SHORT_RELEASE_URL_FORMAT = None

    def set_repo(self, repo):
        """Set repo ID property of project holder instance."""
        self.repo = repo
        self.name = repo.split('/')[-1]

    def __init__(self):
        super(ProjectHolder, self).__init__()

        app_name = __name__.split('.')[0]

        self.cache_dir = user_cache_dir(app_name)
        log.info("Using cache directory: {}.".format(self.cache_dir))
        self.cache = FileCache(self.cache_dir)
        cache_adapter = CacheControlAdapter(cache=self.cache)
        self.mount("http://", cache_adapter)
        self.mount("https://", cache_adapter)

        self.headers.update({'User-Agent': '{}/{}'.format(app_name, __version__)})
        log.info('Created instance of {}'.format(type(self).__name__))
        self.branches = None
        self.only = None
        self.exclude = None
        self.having_asset = None
        self.hostname = None
        # identifies project on a given hostname
        self.repo = None
        # short name for "repo", useful in URLs
        self.name = None
        # in some case we do not specify repo, but feed is discovered, no repo is given then
        self.feed_url = None

    def is_valid(self):
        """Check if project holder is valid instance."""
        return self.feed_url or self.name

    def set_branches(self, branches):
        """Sets project holder's branches."""
        self.branches = branches

    def set_only(self, only):
        """Sets "only" tag selector for this holder."""
        self.only = only
        if only:
            log.info('Only considering tags with "{}"'.format(only))
        return self

    def set_exclude(self, exclude):
        """Sets "exclude" tag selector for this holder."""
        self.exclude = exclude
        if exclude:
            log.info('Only considering tags without "{}"'.format(exclude))
        return self

    def set_having_asset(self, having_asset):
        """Sets "having_asset" selector for this holder."""
        self.having_asset = having_asset
        if having_asset:
            log.info('Only considering releases with asset "{}"'.format(having_asset))
        return self

    @classmethod
    def get_host_repo_for_link(cls, repo):
        """Return hostname and repo from a link."""
        hostname = None
        # return repo modified to result of extraction
        if repo.startswith(('https://', 'http://')):
            # parse hostname for passing to whatever holder selected
            url_parts = repo.split('/')
            hostname = url_parts[2]
            offset = 3 + cls.REPO_URL_PROJECT_OFFSET
            repo = "/".join(url_parts[offset:offset + cls.REPO_URL_PROJECT_COMPONENTS])
        return hostname, repo

    @classmethod
    def is_official_for_repo(cls, repo):
        """Check if repo is a known repo for this type of project holder."""
        if repo.startswith(('https://', 'http://')):
            for url in cls.KNOWN_REPO_URLS:
                if repo.startswith((url, "https://{}".format(url), "http://{}".format(url))):
                    log.info('{} Starts with {}'.format(repo, url))
                    return cls.KNOWN_REPO_URLS[url]
        else:
            if repo.lower() in cls.KNOWN_REPOS_BY_NAME:
                log.info('Selecting known repo {}'.format(repo))
                return cls.KNOWN_REPOS_BY_NAME[repo.lower()]
        return False

    @classmethod
    def get_matching_hostname(cls, repo):
        """Find matching hostname between repo and holder's default hostname."""
        if not cls.DEFAULT_HOSTNAME:
            return None
        if repo.startswith('http://{}'.format(cls.DEFAULT_HOSTNAME)):
            return cls.DEFAULT_HOSTNAME
        if repo.startswith('https://{}'.format(cls.DEFAULT_HOSTNAME)):
            return cls.DEFAULT_HOSTNAME
        return None

    def matches_major_filter(self, version, major):
        if self.branches and major in self.branches and \
                re.search(r"{}".format(self.branches[major]), str(version)):
            log.info('{} matches major {}'.format(version, self.branches[major]))
            return True
        if str(version).startswith('{}.'.format(major)):
            log.info('{} is under the desired major {}'.format(
                version, major))
            return True
        if str(version) == major:
            return True
        return False

    def matches_only(self, version_s):
        if not self.only:
            return True
        search = self.only
        positive = True
        if search.startswith('!'):
            positive = False
            search = search[1:]
        if search.startswith('~'):
            search = r'{}'.format(self.only.lstrip('~'))
            return positive == bool(re.search(search, version_s))
        return positive == bool(search in version_s)

    def sanitize_version(self, version_s, pre_ok=False, major=None):
        """Extract version from tag name."""
        log.info("Sanitizing string {} as a satisfying version.".format(version_s))
        res = False
        if not self.matches_only(version_s):
            log.info('"{}" does not match the "only" constraint "{}"'.format(version_s, self.only))
            return res
        try:
            char_fix_required = self.repo in self.LAST_CHAR_FIX_REQUIRED_ON
            v = Version(version_s, char_fix_required=char_fix_required)
            if not v.is_prerelease or pre_ok:
                log.info("Parsed as Version OK. String representation: {}.".format(v))
                res = v
            else:
                log.info("Parsed as unwanted pre-release version: {}.".format(v))
        except InvalidVersion:
            log.info("Failed to parse {} as Version.".format(version_s))
            # attempt to remove extraneous chars and revalidate
            # we use findall for cases where "tag" may be 'foo/2.x/2.45'
            matches = re.findall(r'([0-9]+([.][0-9x]+)+(rc[0-9]?)?)', version_s)
            for s in matches:
                version_s = s[0]
                log.info("Sanitized tag name value to {}.".format(version_s))
                # 1.10.x is a dev release without clear version, so even pre ok will not get it
                if not version_s.endswith('.x'):
                    # we know regex is valid version format, so no need to try catch
                    res = Version(version_s)
            if not matches:
                log.info("Did not find anything that looks like a version in the tag")
                # as a last resort, let's try to convert underscores to dots, while stripping out
                # any "alphanumeric_". many hg repos do this, e.g. PROJECT_1_2_3
                parts = version_s.split('_')
                if len(parts) >= 2 and parts[0].isalpha():
                    # gets list except first item, joins by dot
                    version_s = '.'.join(parts[1:])
                    try:
                        v = Version(version_s)
                        if not v.is_prerelease or pre_ok:
                            log.info("Parsed as Version OK")
                            log.info("String representation of version is {}.".format(v))
                            res = v
                        else:
                            log.info("Parsed as unwanted pre-release version: {}.".format(v))
                    except InvalidVersion:
                        log.info('Still not a valid version after applying underscores fix')
        # apply --major filter
        if res and major and not self.matches_major_filter(res, major):
            log.info('{} is not under the desired major {}'.format(
                version_s, major))
            res = False
        return res

    def _type(self):
        """Get project holder's class name."""
        return self.__class__.__name__

    def release_download_url(self, release, shorter=False):
        """Get release download URL."""
        if not self.RELEASE_URL_FORMAT:
            raise NotImplementedError(
                'Getting release URL for {} is not implemented'.format(self._type()))
        ext = 'zip' if os.name == 'nt' else 'tar.gz'

        fmt = self.SHORT_RELEASE_URL_FORMAT if shorter and self.SHORT_RELEASE_URL_FORMAT else \
            self.RELEASE_URL_FORMAT

        return fmt.format(
            hostname=self.hostname,
            repo=self.repo,
            name=self.name,
            tag=release['tag_name'],
            ext=ext,
            version=release['version']
        )

    def get_assets(self, release, short_urls, assets_filter=None):
        urls = []
        if 'assets' in release and release['assets']:
            for asset in release['assets']:
                if assets_filter and not re.search(assets_filter, asset['name']):
                    continue
                if not assets_filter and asset_does_not_belong_to_machine(asset['name']):
                    continue
                urls.append(asset['browser_download_url'])
        else:
            download_url = self.release_download_url(release, short_urls)
            if not assets_filter or re.search(assets_filter, download_url):
                urls.append(download_url)
        return urls

    def get_canonical_link(self):
        if self.feed_url:
            return self.feed_url
        return 'https://{}/{}'.format(self.hostname, self.repo)



# -*- coding: utf-8 -*-

from dateutil import parser
from bs4 import BeautifulSoup
import logging
from .ProjectHolder import ProjectHolder

log = logging.getLogger(__name__)


def remove_words(title):
    parts = title.split(' ')
    parts_n = []
    for part in parts:
        if not part.isalpha() or '.post' in part:
            parts_n.append(part)
    return " ".join(parts_n)


class WikipediaRepoSession(ProjectHolder):
    KNOWN_REPOS_BY_NAME = {
        'alpine': {
            'repo': 'Alpine_Linux',
        },
        'rocky': {
            'repo': 'Rocky_Linux',
        },
        'rockylinux': {
            'repo': 'Rocky_Linux',
        },
        'fedora': {
            'repo': 'Fedora_(operating_system)'
        },
        'rhel': {
            'repo': 'Red_Hat_Enterprise_Linux'
        },
        'redhat': {
            'repo': 'Red_Hat_Enterprise_Linux'
        },
        'almalinux': {
            'repo': 'AlmaLinux'
        },
        'ios': {
            'repo': 'IOS'
        },
        'ubuntu': {
            'repo': 'Ubuntu'
        },
        'debian': {
            'repo': 'Debian'
        },
        'android': {
            'repo': 'Android_(operating_system)'
        },
        'windows': {
            'repo': 'Microsoft_Windows'
        },
        'osx': {
            'repo': 'MacOS'
        },
        'sles': {
            'repo': 'SUSE_Linux_Enterprise'
        },
        'opensuse': {
            'repo': 'OpenSUSE'
        }
    }

    REPO_URL_PROJECT_COMPONENTS = 1
    DEFAULT_HOSTNAME = 'en.wikipedia.org'
    # For project URLs, e.g. https://en.wikipedia.org/wiki/Rocky_Linux
    # a URI does not start with a repo name, skip '/wiki/'
    REPO_URL_PROJECT_OFFSET = 1

    def __init__(self, repo, hostname):
        super(WikipediaRepoSession, self).__init__()
        self.hostname = hostname
        if not self.hostname:
            self.hostname = self.DEFAULT_HOSTNAME
        self.set_repo(repo)

    def get_latest(self, pre_ok=False, major=None):
        ret = None
        tag_name = None
        tag = {}
        r = self.get('https://{}/wiki/{}'.format(self.hostname, self.repo))
        soup = BeautifulSoup(r.text, "html.parser")
        # we only need the first one
        infobox = soup.select_one(".infobox")
        links = infobox.select("a")
        for link in links:
            if link.text.lower() in ['latest release', 'stable release']:
                release_data = link.parent.parent.select_one(".infobox-data")
                # get published before it's removed:
                published = release_data.select_one('span.published').text
                published = parser.parse(published)
                for t in release_data.select('sup, span'):
                    t.decompose()
                tag_name = release_data.text
                tag_name = tag_name.replace(' Service Pack ', '.post')
                # remove alphas from beginning
                tag_name = remove_words(tag_name).split('/')[0]
                # Remove unicode stuff (for Python 2)
                tag['title'] = release_data.text.encode("ascii", "ignore").decode()
                log.info('Pre-parsed title: {}'.format(tag['title']))
                tag['tag_date'] = published
                break
        if not tag_name:
            return None
        # Remove unicode stuff (for Python 2)
        tag_name = tag_name.encode("ascii", "ignore").decode()
        version = self.sanitize_version(tag_name, pre_ok, major)
        if not version:
            return None
        if not ret or version > ret['version']:
            tag['tag_name'] = tag_name
            tag['version'] = version
            ret = tag
        return ret
